@@ -40,6 +40,127 @@
 #define DRI_CONF_VBLANK_DEF_INTERVAL_1 2
 #define DRI_CONF_VBLANK_ALWAYS_SYNC 3
 
+/**
+ * A cached blit context.
+ */
+struct loader_dri3_blit_context {
+   mtx_t mtx;
+   __DRIcontext *ctx;
+   __DRIscreen *cur_screen;
+   const __DRIcoreExtension *core;
+};
+
+/* For simplicity we maintain the cache only for a single screen at a time */
+static struct loader_dri3_blit_context blit_context = {
+   _MTX_INITIALIZER_NP, NULL
+};
+
+static void
+dri3_flush_present_events(struct loader_dri3_drawable *draw);
+
+static struct loader_dri3_buffer *
+dri3_find_back_alloc(struct loader_dri3_drawable *draw);
+
+/**
+ * Do we have blit functionality in the image blit extension?
+ *
+ * \param draw[in]  The drawable intended to blit from / to.
+ * \return  true if we have blit functionality. false otherwise.
+ */
+static bool loader_dri3_have_image_blit(const struct loader_dri3_drawable *draw)
+{
+   return draw->ext->image->base.version >= 9 &&
+      draw->ext->image->blitImage != NULL;
+}
+
+/**
+ * Get and lock (for use with the current thread) a dri context associated
+ * with the drawable's dri screen. The context is intended to be used with
+ * the dri image extension's blitImage method.
+ *
+ * \param draw[in]  Pointer to the drawable whose dri screen we want a
+ * dri context for.
+ * \return A dri context or NULL if context creation failed.
+ *
+ * When the caller is done with the context (even if the context returned was
+ * NULL), the caller must call loader_dri3_blit_context_put.
+ */
+static __DRIcontext *
+loader_dri3_blit_context_get(struct loader_dri3_drawable *draw)
+{
+   mtx_lock(&blit_context.mtx);
+
+   if (blit_context.ctx && blit_context.cur_screen != draw->dri_screen) {
+      blit_context.core->destroyContext(blit_context.ctx);
+      blit_context.ctx = NULL;
+   }
+
+   if (!blit_context.ctx) {
+      blit_context.ctx = draw->ext->core->createNewContext(draw->dri_screen,
+                                                           NULL, NULL, NULL);
+      blit_context.cur_screen = draw->dri_screen;
+      blit_context.core = draw->ext->core;
+   }
+
+   return blit_context.ctx;
+}
+
+/**
+ * Release (for use with other threads) a dri context previously obtained using
+ * loader_dri3_blit_context_get.
+ */
+static void
+loader_dri3_blit_context_put(void)
+{
+   mtx_unlock(&blit_context.mtx);
+}
+
+/**
+ * Blit (parts of) the contents of a DRI image to another dri image
+ *
+ * \param draw[in]  The drawable which owns the images.
+ * \param dst[in]  The destination image.
+ * \param src[in]  The source image.
+ * \param dstx0[in]  Start destination coordinate.
+ * \param dsty0[in]  Start destination coordinate.
+ * \param width[in]  Blit width.
+ * \param height[in] Blit height.
+ * \param srcx0[in]  Start source coordinate.
+ * \param srcy0[in]  Start source coordinate.
+ * \param flush_flag[in]  Image blit flush flag.
+ * \return true iff successful.
+ */
+static bool
+loader_dri3_blit_image(struct loader_dri3_drawable *draw,
+                       __DRIimage *dst, __DRIimage *src,
+                       int dstx0, int dsty0, int width, int height,
+                       int srcx0, int srcy0, int flush_flag)
+{
+   __DRIcontext *dri_context;
+   bool use_blit_context = false;
+
+   if (!loader_dri3_have_image_blit(draw))
+      return false;
+
+   dri_context = draw->vtable->get_dri_context(draw);
+
+   if (!dri_context || !draw->vtable->in_current_context(draw)) {
+      dri_context = loader_dri3_blit_context_get(draw);
+      use_blit_context = true;
+      flush_flag |= __BLIT_FLAG_FLUSH;
+   }
+
+   if (dri_context)
+      draw->ext->image->blitImage(dri_context, dst, src, dstx0, dsty0,
+                                  width, height, srcx0, srcy0,
+                                  width, height, flush_flag);
+
+   if (use_blit_context)
+      loader_dri3_blit_context_put();
+
+   return dri_context != NULL;
+}
+
 static inline void
 dri3_fence_reset(xcb_connection_t *c, struct loader_dri3_buffer *buffer)
 {
@@ -59,10 +180,16 @@ dri3_fence_trigger(xcb_connection_t *c, struct loader_dri3_buffer *buffer)
 }
 
 static inline void
-dri3_fence_await(xcb_connection_t *c, struct loader_dri3_buffer *buffer)
+dri3_fence_await(xcb_connection_t *c, struct loader_dri3_drawable *draw,
+                 struct loader_dri3_buffer *buffer)
 {
    xcb_flush(c);
    xshmfence_await(buffer->shm_fence);
+   if (draw) {
+      mtx_lock(&draw->mtx);
+      dri3_flush_present_events(draw);
+      mtx_unlock(&draw->mtx);
+   }
 }
 
 static void
@@ -77,9 +204,7 @@ dri3_update_num_back(struct loader_dri3_drawable *draw)
 void
 loader_dri3_set_swap_interval(struct loader_dri3_drawable *draw, int interval)
 {
-   interval = draw->vtable->clamp_swap_interval(draw, interval);
-   draw->vtable->set_swap_interval(draw, interval);
-   dri3_update_num_back(draw);
+   draw->swap_interval = interval;
 }
 
 /** dri3_free_render_buffer
@@ -121,6 +246,9 @@ loader_dri3_drawable_fini(struct loader_dri3_drawable *draw)
       xcb_discard_reply(draw->conn, cookie.sequence);
       xcb_unregister_for_special_event(draw->conn, draw->special_event);
    }
+
+   cnd_destroy(&draw->event_cnd);
+   mtx_destroy(&draw->mtx);
 }
 
 int
@@ -150,6 +278,11 @@ loader_dri3_drawable_init(xcb_connection_t *conn,
    draw->have_fake_front = 0;
    draw->first_init = true;
 
+   draw->cur_blit_source = -1;
+   draw->back_format = __DRI_IMAGE_FORMAT_NONE;
+   mtx_init(&draw->mtx, mtx_plain);
+   cnd_init(&draw->event_cnd);
+
    if (draw->ext->config)
       draw->ext->config->configQueryi(draw->dri_screen,
                                       "vblank_mode", &vblank_mode);
@@ -165,7 +298,7 @@ loader_dri3_drawable_init(xcb_connection_t *conn,
       swap_interval = 1;
       break;
    }
-   draw->vtable->set_swap_interval(draw, swap_interval);
+   draw->swap_interval = swap_interval;
 
    dri3_update_num_back(draw);
 
@@ -191,6 +324,13 @@ loader_dri3_drawable_init(xcb_connection_t *conn,
    draw->vtable->set_drawable_size(draw, draw->width, draw->height);
    free(reply);
 
+   draw->swap_method = __DRI_ATTRIB_SWAP_UNDEFINED;
+   if (draw->ext->core->base.version >= 2) {
+      (void )draw->ext->core->getConfigAttrib(dri_config,
+                                              __DRI_ATTRIB_SWAP_METHOD,
+                                              &draw->swap_method);
+   }
+
    /*
     * Make sure server has the same swap interval we do for the new
     * drawable.
@@ -214,6 +354,7 @@ dri3_handle_present_event(struct loader_dri3_drawable *draw,
       draw->width = ce->width;
       draw->height = ce->height;
       draw->vtable->set_drawable_size(draw, draw->width, draw->height);
+      draw->ext->flush->invalidate(draw->dri_drawable);
       break;
    }
    case XCB_PRESENT_COMPLETE_NOTIFY: {
@@ -235,7 +376,6 @@ dri3_handle_present_event(struct loader_dri3_drawable *draw,
             draw->flipping = false;
             break;
          }
-         dri3_update_num_back(draw);
 
          if (draw->vtable->show_fps)
             draw->vtable->show_fps(draw, ce->ust);
@@ -256,13 +396,14 @@ dri3_handle_present_event(struct loader_dri3_drawable *draw,
       for (b = 0; b < sizeof(draw->buffers) / sizeof(draw->buffers[0]); b++) {
          struct loader_dri3_buffer *buf = draw->buffers[b];
 
-         if (buf && buf->pixmap == ie->pixmap) {
+         if (buf && buf->pixmap == ie->pixmap)
             buf->busy = 0;
-            if (draw->num_back <= b && b < LOADER_DRI3_MAX_BACK) {
-               dri3_free_render_buffer(draw, buf);
-               draw->buffers[b] = NULL;
-            }
-            break;
+
+         if (buf && draw->num_back <= b && b < LOADER_DRI3_MAX_BACK &&
+             draw->cur_blit_source != b &&
+             !buf->busy) {
+            dri3_free_render_buffer(draw, buf);
+            draw->buffers[b] = NULL;
          }
       }
       break;
@@ -272,13 +413,27 @@ dri3_handle_present_event(struct loader_dri3_drawable *draw,
 }
 
 static bool
-dri3_wait_for_event(struct loader_dri3_drawable *draw)
+dri3_wait_for_event_locked(struct loader_dri3_drawable *draw)
 {
    xcb_generic_event_t *ev;
    xcb_present_generic_event_t *ge;
 
    xcb_flush(draw->conn);
-   ev = xcb_wait_for_special_event(draw->conn, draw->special_event);
+
+   /* Only have one thread waiting for events at a time */
+   if (draw->has_event_waiter) {
+      cnd_wait(&draw->event_cnd, &draw->mtx);
+      /* Another thread has updated the protected info, so retest. */
+      return true;
+   } else {
+      draw->has_event_waiter = true;
+      /* Allow other threads access to the drawable while we're waiting. */
+      mtx_unlock(&draw->mtx);
+      ev = xcb_wait_for_special_event(draw->conn, draw->special_event);
+      mtx_lock(&draw->mtx);
+      draw->has_event_waiter = false;
+      cnd_broadcast(&draw->event_cnd);
+   }
    if (!ev)
       return false;
    ge = (void *) ev;
@@ -307,19 +462,23 @@ loader_dri3_wait_for_msc(struct loader_dri3_drawable *draw,
                           divisor,
                           remainder);
 
+   mtx_lock(&draw->mtx);
    xcb_flush(draw->conn);
 
    /* Wait for the event */
    if (draw->special_event) {
       while ((int32_t) (msc_serial - draw->recv_msc_serial) > 0) {
-         if (!dri3_wait_for_event(draw))
+         if (!dri3_wait_for_event_locked(draw)) {
+            mtx_unlock(&draw->mtx);
             return false;
+         }
       }
    }
 
    *ust = draw->notify_ust;
    *msc = draw->notify_msc;
    *sbc = draw->recv_sbc;
+   mtx_unlock(&draw->mtx);
 
    return true;
 }
@@ -341,17 +500,21 @@ loader_dri3_wait_for_sbc(struct loader_dri3_drawable *draw,
     *      swaps requested with glXSwapBuffersMscOML for that window have
     *      completed."
     */
+   mtx_lock(&draw->mtx);
    if (!target_sbc)
       target_sbc = draw->send_sbc;
 
    while (draw->recv_sbc < target_sbc) {
-      if (!dri3_wait_for_event(draw))
+      if (!dri3_wait_for_event_locked(draw)) {
+         mtx_unlock(&draw->mtx);
          return 0;
+      }
    }
 
    *ust = draw->ust;
    *msc = draw->msc;
    *sbc = draw->recv_sbc;
+   mtx_unlock(&draw->mtx);
    return 1;
 }
 
@@ -364,25 +527,37 @@ static int
 dri3_find_back(struct loader_dri3_drawable *draw)
 {
    int b;
-   xcb_generic_event_t *ev;
-   xcb_present_generic_event_t *ge;
+   int num_to_consider;
+
+   mtx_lock(&draw->mtx);
+   /* Increase the likelyhood of reusing current buffer */
+   dri3_flush_present_events(draw);
+
+   /* Check whether we need to reuse the current back buffer as new back.
+    * In that case, wait until it's not busy anymore.
+    */
+   dri3_update_num_back(draw);
+   num_to_consider = draw->num_back;
+   if (!loader_dri3_have_image_blit(draw) && draw->cur_blit_source != -1) {
+      num_to_consider = 1;
+      draw->cur_blit_source = -1;
+   }
 
    for (;;) {
-      for (b = 0; b < draw->num_back; b++) {
+      for (b = 0; b < num_to_consider; b++) {
          int id = LOADER_DRI3_BACK_ID((b + draw->cur_back) % draw->num_back);
          struct loader_dri3_buffer *buffer = draw->buffers[id];
 
          if (!buffer || !buffer->busy) {
             draw->cur_back = id;
+            mtx_unlock(&draw->mtx);
             return id;
          }
       }
-      xcb_flush(draw->conn);
-      ev = xcb_wait_for_special_event(draw->conn, draw->special_event);
-      if (!ev)
+      if (!dri3_wait_for_event_locked(draw)) {
+         mtx_unlock(&draw->mtx);
          return -1;
-      ge = (void *) ev;
-      dri3_handle_present_event(draw, ge);
+      }
    }
 }
 
@@ -468,9 +643,6 @@ loader_dri3_copy_sub_buffer(struct loader_dri3_drawable *draw,
 {
    struct loader_dri3_buffer *back;
    unsigned flags = __DRI2_FLUSH_DRAWABLE;
-   __DRIcontext *dri_context;
-
-   dri_context = draw->vtable->get_dri_context(draw);
 
    /* Check we have the right attachments */
    if (!draw->have_back || draw->is_pixmap)
@@ -480,34 +652,27 @@ loader_dri3_copy_sub_buffer(struct loader_dri3_drawable *draw,
       flags |= __DRI2_FLUSH_CONTEXT;
    loader_dri3_flush(draw, flags, __DRI2_THROTTLE_SWAPBUFFER);
 
-   back = dri3_back_buffer(draw);
+   back = dri3_find_back_alloc(draw);
+   if (!back)
+      return;
+
    y = draw->height - y - height;
 
-   if (draw->is_different_gpu && draw->vtable->in_current_context(draw)) {
+   if (draw->is_different_gpu) {
       /* Update the linear buffer part of the back buffer
        * for the dri3_copy_area operation
        */
-      draw->ext->image->blitImage(dri_context,
-                                  back->linear_buffer,
-                                  back->image,
-                                  0, 0, back->width,
-                                  back->height,
-                                  0, 0, back->width,
-                                  back->height, __BLIT_FLAG_FLUSH);
-      /* We use blitImage to update our fake front,
-       */
-      if (draw->have_fake_front)
-         draw->ext->image->blitImage(dri_context,
-                                     dri3_fake_front_buffer(draw)->image,
-                                     back->image,
-                                     x, y, width, height,
-                                     x, y, width, height, __BLIT_FLAG_FLUSH);
+      (void) loader_dri3_blit_image(draw,
+                                    back->linear_buffer,
+                                    back->image,
+                                    0, 0, back->width, back->height,
+                                    0, 0, __BLIT_FLAG_FLUSH);
    }
 
    loader_dri3_swapbuffer_barrier(draw);
    dri3_fence_reset(draw->conn, back);
    dri3_copy_area(draw->conn,
-                  dri3_back_buffer(draw)->pixmap,
+                  back->pixmap,
                   draw->drawable,
                   dri3_drawable_gc(draw),
                   x, y, x, y, width, height);
@@ -515,17 +680,23 @@ loader_dri3_copy_sub_buffer(struct loader_dri3_drawable *draw,
    /* Refresh the fake front (if present) after we just damaged the real
     * front.
     */
-   if (draw->have_fake_front && !draw->is_different_gpu) {
+   if (draw->have_fake_front &&
+       !loader_dri3_blit_image(draw,
+                               dri3_fake_front_buffer(draw)->image,
+                               back->image,
+                               x, y, width, height,
+                               x, y, __BLIT_FLAG_FLUSH) &&
+       !draw->is_different_gpu) {
       dri3_fence_reset(draw->conn, dri3_fake_front_buffer(draw));
       dri3_copy_area(draw->conn,
-                     dri3_back_buffer(draw)->pixmap,
+                     back->pixmap,
                      dri3_fake_front_buffer(draw)->pixmap,
                      dri3_drawable_gc(draw),
                      x, y, x, y, width, height);
       dri3_fence_trigger(draw->conn, dri3_fake_front_buffer(draw));
-      dri3_fence_await(draw->conn, dri3_fake_front_buffer(draw));
+      dri3_fence_await(draw->conn, NULL, dri3_fake_front_buffer(draw));
    }
-   dri3_fence_await(draw->conn, back);
+   dri3_fence_await(draw->conn, draw, back);
 }
 
 void
@@ -541,20 +712,18 @@ loader_dri3_copy_drawable(struct loader_dri3_drawable *draw,
                   dri3_drawable_gc(draw),
                   0, 0, 0, 0, draw->width, draw->height);
    dri3_fence_trigger(draw->conn, dri3_fake_front_buffer(draw));
-   dri3_fence_await(draw->conn, dri3_fake_front_buffer(draw));
+   dri3_fence_await(draw->conn, draw, dri3_fake_front_buffer(draw));
 }
 
 void
 loader_dri3_wait_x(struct loader_dri3_drawable *draw)
 {
    struct loader_dri3_buffer *front;
-   __DRIcontext *dri_context;
 
    if (draw == NULL || !draw->have_fake_front)
       return;
 
    front = dri3_fake_front_buffer(draw);
-   dri_context = draw->vtable->get_dri_context(draw);
 
    loader_dri3_copy_drawable(draw, front->pixmap, draw->drawable);
 
@@ -563,39 +732,33 @@ loader_dri3_wait_x(struct loader_dri3_drawable *draw)
     * Copy back to the tiled buffer we use for rendering.
     * Note that we don't need flushing.
     */
-   if (draw->is_different_gpu && draw->vtable->in_current_context(draw))
-      draw->ext->image->blitImage(dri_context,
-                                  front->image,
-                                  front->linear_buffer,
-                                  0, 0, front->width,
-                                  front->height,
-                                  0, 0, front->width,
-                                  front->height, 0);
+   if (draw->is_different_gpu)
+      (void) loader_dri3_blit_image(draw,
+                                    front->image,
+                                    front->linear_buffer,
+                                    0, 0, front->width, front->height,
+                                    0, 0, 0);
 }
 
 void
 loader_dri3_wait_gl(struct loader_dri3_drawable *draw)
 {
    struct loader_dri3_buffer *front;
-   __DRIcontext *dri_context;
 
    if (draw == NULL || !draw->have_fake_front)
       return;
 
    front = dri3_fake_front_buffer(draw);
-   dri_context = draw->vtable->get_dri_context(draw);
 
    /* In the psc->is_different_gpu case, we update the linear_buffer
     * before updating the real front.
     */
-   if (draw->is_different_gpu && draw->vtable->in_current_context(draw))
-      draw->ext->image->blitImage(dri_context,
-                                  front->linear_buffer,
-                                  front->image,
-                                  0, 0, front->width,
-                                  front->height,
-                                  0, 0, front->width,
-                                  front->height, __BLIT_FLAG_FLUSH);
+   if (draw->is_different_gpu)
+      (void) loader_dri3_blit_image(draw,
+                                    front->linear_buffer,
+                                    front->image,
+                                    0, 0, front->width, front->height,
+                                    0, 0, __BLIT_FLAG_FLUSH);
    loader_dri3_swapbuffer_barrier(draw);
    loader_dri3_copy_drawable(draw, draw->drawable, front->pixmap);
 }
@@ -610,6 +773,9 @@ dri3_flush_present_events(struct loader_dri3_drawable *draw)
    /* Check to see if any configuration changes have occurred
     * since we were last invoked
     */
+   if (draw->has_event_waiter)
+      return;
+
    if (draw->special_event) {
       xcb_generic_event_t    *ev;
 
@@ -632,34 +798,42 @@ loader_dri3_swap_buffers_msc(struct loader_dri3_drawable *draw,
                              bool force_copy)
 {
    struct loader_dri3_buffer *back;
-   __DRIcontext *dri_context;
    int64_t ret = 0;
    uint32_t options = XCB_PRESENT_OPTION_NONE;
-   int swap_interval;
-
-   dri_context = draw->vtable->get_dri_context(draw);
-   swap_interval = draw->vtable->get_swap_interval(draw);
 
    draw->vtable->flush_drawable(draw, flush_flags);
 
-   back = draw->buffers[dri3_find_back(draw)];
+   back = dri3_find_back_alloc(draw);
+
+   mtx_lock(&draw->mtx);
    if (draw->is_different_gpu && back) {
       /* Update the linear buffer before presenting the pixmap */
-      draw->ext->image->blitImage(dri_context,
-                                  back->linear_buffer,
-                                  back->image,
-                                  0, 0, back->width,
-                                  back->height,
-                                  0, 0, back->width,
-                                  back->height, __BLIT_FLAG_FLUSH);
-      /* Update the fake front */
-      if (draw->have_fake_front)
-         draw->ext->image->blitImage(dri_context,
-                                     draw->buffers[LOADER_DRI3_FRONT_ID]->image,
-                                     back->image,
-                                     0, 0, draw->width, draw->height,
-                                     0, 0, draw->width, draw->height,
-                                     __BLIT_FLAG_FLUSH);
+      (void) loader_dri3_blit_image(draw,
+                                    back->linear_buffer,
+                                    back->image,
+                                    0, 0, back->width, back->height,
+                                    0, 0, __BLIT_FLAG_FLUSH);
+   }
+
+   /* If we need to preload the new back buffer, remember the source.
+    * The force_copy parameter is used by EGL to attempt to preserve
+    * the back buffer across a call to this function.
+    */
+   if (draw->swap_method != __DRI_ATTRIB_SWAP_UNDEFINED || force_copy)
+      draw->cur_blit_source = LOADER_DRI3_BACK_ID(draw->cur_back);
+
+   /* Exchange the back and fake front. Even though the server knows about these
+    * buffers, it has no notion of back and fake front.
+    */
+   if (back && draw->have_fake_front) {
+      struct loader_dri3_buffer *tmp;
+
+      tmp = dri3_fake_front_buffer(draw);
+      draw->buffers[LOADER_DRI3_FRONT_ID] = back;
+      draw->buffers[LOADER_DRI3_BACK_ID(draw->cur_back)] = tmp;
+
+      if (draw->swap_method == __DRI_ATTRIB_SWAP_COPY  || force_copy)
+         draw->cur_blit_source = LOADER_DRI3_FRONT_ID;
    }
 
    dri3_flush_present_events(draw);
@@ -674,7 +848,7 @@ loader_dri3_swap_buffers_msc(struct loader_dri3_drawable *draw,
        */
       ++draw->send_sbc;
       if (target_msc == 0 && divisor == 0 && remainder == 0)
-         target_msc = draw->msc + swap_interval *
+         target_msc = draw->msc + draw->swap_interval *
                       (draw->send_sbc - draw->recv_sbc);
       else if (divisor == 0 && remainder > 0) {
          /* From the GLX_OML_sync_control spec:
@@ -698,10 +872,15 @@ loader_dri3_swap_buffers_msc(struct loader_dri3_drawable *draw,
        * behaviour by not using XCB_PRESENT_OPTION_ASYNC, but this should not be
        * the default.
        */
-      if (swap_interval == 0)
+      if (draw->swap_interval == 0)
           options |= XCB_PRESENT_OPTION_ASYNC;
-      if (force_copy)
-          options |= XCB_PRESENT_OPTION_COPY;
+
+      /* If we need to populate the new back, but need to reuse the back
+       * buffer slot due to lack of local blit capabilities, make sure
+       * the server doesn't flip and we deadlock.
+       */
+      if (!loader_dri3_have_image_blit(draw) && draw->cur_blit_source != -1)
+         options |= XCB_PRESENT_OPTION_COPY;
 
       back->busy = 1;
       back->last_swap = draw->send_sbc;
@@ -722,25 +901,31 @@ loader_dri3_swap_buffers_msc(struct loader_dri3_drawable *draw,
                          remainder, 0, NULL);
       ret = (int64_t) draw->send_sbc;
 
-      /* If there's a fake front, then copy the source back buffer
-       * to the fake front to keep it up to date. This needs
-       * to reset the fence and make future users block until
-       * the X server is done copying the bits
+      /* Schedule a server-side back-preserving blit if necessary.
+       * This happens iff all conditions below are satisfied:
+       * a) We have a fake front,
+       * b) We need to preserve the back buffer,
+       * c) We don't have local blit capabilities.
        */
-      if (draw->have_fake_front && !draw->is_different_gpu) {
-         dri3_fence_reset(draw->conn, draw->buffers[LOADER_DRI3_FRONT_ID]);
-         dri3_copy_area(draw->conn,
-                        back->pixmap,
-                        draw->buffers[LOADER_DRI3_FRONT_ID]->pixmap,
+      if (!loader_dri3_have_image_blit(draw) && draw->cur_blit_source != -1 &&
+          draw->cur_blit_source != LOADER_DRI3_BACK_ID(draw->cur_back)) {
+         struct loader_dri3_buffer *new_back = dri3_back_buffer(draw);
+         struct loader_dri3_buffer *src = draw->buffers[draw->cur_blit_source];
+
+         dri3_fence_reset(draw->conn, new_back);
+         dri3_copy_area(draw->conn, src->pixmap,
+                        new_back->pixmap,
                         dri3_drawable_gc(draw),
-                        0, 0, 0, 0,
-                        draw->width, draw->height);
-         dri3_fence_trigger(draw->conn, draw->buffers[LOADER_DRI3_FRONT_ID]);
+                        0, 0, 0, 0, draw->width, draw->height);
+         dri3_fence_trigger(draw->conn, new_back);
+         new_back->last_swap = src->last_swap;
       }
+
       xcb_flush(draw->conn);
       if (draw->stamp)
          ++(*draw->stamp);
    }
+   mtx_unlock(&draw->mtx);
 
    draw->ext->flush->invalidate(draw->dri_drawable);
 
@@ -750,15 +935,15 @@ loader_dri3_swap_buffers_msc(struct loader_dri3_drawable *draw,
 int
 loader_dri3_query_buffer_age(struct loader_dri3_drawable *draw)
 {
-   int back_id = LOADER_DRI3_BACK_ID(dri3_find_back(draw));
+   struct loader_dri3_buffer *back = dri3_find_back_alloc(draw);
+   int ret;
 
-   if (back_id < 0 || !draw->buffers[back_id])
-      return 0;
+   mtx_lock(&draw->mtx);
+   ret = (!back || back->last_swap == 0) ? 0 :
+      draw->send_sbc - back->last_swap + 1;
+   mtx_unlock(&draw->mtx);
 
-   if (draw->buffers[back_id]->last_swap != 0)
-      return draw->send_sbc - draw->buffers[back_id]->last_swap + 1;
-   else
-      return 0;
+   return ret;
 }
 
 /** loader_dri3_open
@@ -816,6 +1001,27 @@ dri3_cpp_for_format(uint32_t format) {
    }
 }
 
+/* the DRIimage createImage function takes __DRI_IMAGE_FORMAT codes, while
+ * the createImageFromFds call takes __DRI_IMAGE_FOURCC codes. To avoid
+ * complete confusion, just deal in __DRI_IMAGE_FORMAT codes for now and
+ * translate to __DRI_IMAGE_FOURCC codes in the call to createImageFromFds
+ */
+static int
+image_format_to_fourcc(int format)
+{
+
+   /* Convert from __DRI_IMAGE_FORMAT to __DRI_IMAGE_FOURCC (sigh) */
+   switch (format) {
+   case __DRI_IMAGE_FORMAT_SARGB8: return __DRI_IMAGE_FOURCC_SARGB8888;
+   case __DRI_IMAGE_FORMAT_RGB565: return __DRI_IMAGE_FOURCC_RGB565;
+   case __DRI_IMAGE_FORMAT_XRGB8888: return __DRI_IMAGE_FOURCC_XRGB8888;
+   case __DRI_IMAGE_FORMAT_ARGB8888: return __DRI_IMAGE_FOURCC_ARGB8888;
+   case __DRI_IMAGE_FORMAT_ABGR8888: return __DRI_IMAGE_FOURCC_ABGR8888;
+   case __DRI_IMAGE_FORMAT_XBGR8888: return __DRI_IMAGE_FOURCC_XBGR8888;
+   }
+   return 0;
+}
+
 /** loader_dri3_alloc_render_buffer
  *
  * Use the driver createImage function to construct a __DRIimage, then
@@ -955,6 +1161,7 @@ static int
 dri3_update_drawable(__DRIdrawable *driDrawable,
                      struct loader_dri3_drawable *draw)
 {
+   mtx_lock(&draw->mtx);
    if (draw->first_init) {
       xcb_get_geometry_cookie_t                 geom_cookie;
       xcb_get_geometry_reply_t                  *geom_reply;
@@ -995,8 +1202,10 @@ dri3_update_drawable(__DRIdrawable *driDrawable,
 
       geom_reply = xcb_get_geometry_reply(draw->conn, geom_cookie, NULL);
 
-      if (!geom_reply)
+      if (!geom_reply) {
+         mtx_unlock(&draw->mtx);
          return false;
+      }
 
       draw->width = geom_reply->width;
       draw->height = geom_reply->height;
@@ -1028,6 +1237,7 @@ dri3_update_drawable(__DRIdrawable *driDrawable,
       if (error) {
          if (error->error_code != BadWindow) {
             free(error);
+            mtx_unlock(&draw->mtx);
             return false;
          }
          draw->is_pixmap = true;
@@ -1036,30 +1246,10 @@ dri3_update_drawable(__DRIdrawable *driDrawable,
       }
    }
    dri3_flush_present_events(draw);
+   mtx_unlock(&draw->mtx);
    return true;
 }
 
-/* the DRIimage createImage function takes __DRI_IMAGE_FORMAT codes, while
- * the createImageFromFds call takes __DRI_IMAGE_FOURCC codes. To avoid
- * complete confusion, just deal in __DRI_IMAGE_FORMAT codes for now and
- * translate to __DRI_IMAGE_FOURCC codes in the call to createImageFromFds
- */
-static int
-image_format_to_fourcc(int format)
-{
-
-   /* Convert from __DRI_IMAGE_FORMAT to __DRI_IMAGE_FOURCC (sigh) */
-   switch (format) {
-   case __DRI_IMAGE_FORMAT_SARGB8: return __DRI_IMAGE_FOURCC_SARGB8888;
-   case __DRI_IMAGE_FORMAT_RGB565: return __DRI_IMAGE_FOURCC_RGB565;
-   case __DRI_IMAGE_FORMAT_XRGB8888: return __DRI_IMAGE_FOURCC_XRGB8888;
-   case __DRI_IMAGE_FORMAT_ARGB8888: return __DRI_IMAGE_FOURCC_ARGB8888;
-   case __DRI_IMAGE_FORMAT_ABGR8888: return __DRI_IMAGE_FOURCC_ABGR8888;
-   case __DRI_IMAGE_FORMAT_XBGR8888: return __DRI_IMAGE_FOURCC_XBGR8888;
-   }
-   return 0;
-}
-
 __DRIimage *
 loader_dri3_create_image(xcb_connection_t *c,
                          xcb_dri3_buffer_from_pixmap_reply_t *bp_reply,
@@ -1119,7 +1309,6 @@ dri3_get_pixmap_buffer(__DRIdrawable *driDrawable, unsigned int format,
    xcb_sync_fence_t                     sync_fence;
    struct xshmfence                     *shm_fence;
    int                                  fence_fd;
-   __DRIscreen                          *cur_screen;
 
    if (buffer)
       return buffer;
@@ -1150,17 +1339,8 @@ dri3_get_pixmap_buffer(__DRIdrawable *driDrawable, unsigned int format,
    if (!bp_reply)
       goto no_image;
 
-   /* Get the currently-bound screen or revert to using the drawable's screen if
-    * no contexts are currently bound. The latter case is at least necessary for
-    * obs-studio, when using Window Capture (Xcomposite) as a Source.
-    */
-   cur_screen = draw->vtable->get_dri_screen(draw);
-   if (!cur_screen) {
-       cur_screen = draw->dri_screen;
-   }
-
    buffer->image = loader_dri3_create_image(draw->conn, bp_reply, format,
-                                            cur_screen, draw->ext->image,
+                                            draw->dri_screen, draw->ext->image,
                                             buffer);
    if (!buffer->image)
       goto no_image;
@@ -1169,7 +1349,6 @@ dri3_get_pixmap_buffer(__DRIdrawable *driDrawable, unsigned int format,
    buffer->own_pixmap = false;
    buffer->width = bp_reply->width;
    buffer->height = bp_reply->height;
-   buffer->buffer_type = buffer_type;
    buffer->shm_fence = shm_fence;
    buffer->sync_fence = sync_fence;
 
@@ -1201,11 +1380,10 @@ dri3_get_buffer(__DRIdrawable *driDrawable,
 {
    struct loader_dri3_buffer *buffer;
    int buf_id;
-   __DRIcontext *dri_context;
-
-   dri_context = draw->vtable->get_dri_context(draw);
 
    if (buffer_type == loader_dri3_buffer_back) {
+      draw->back_format = format;
+
       buf_id = dri3_find_back(draw);
 
       if (buf_id < 0)
@@ -1236,30 +1414,30 @@ dri3_get_buffer(__DRIdrawable *driDrawable,
       /* When resizing, copy the contents of the old buffer, waiting for that
        * copy to complete using our fences before proceeding
        */
-      switch (buffer_type) {
-      case loader_dri3_buffer_back:
-         if (buffer) {
-            if (!buffer->linear_buffer) {
-               dri3_fence_reset(draw->conn, new_buffer);
-               dri3_fence_await(draw->conn, buffer);
-               dri3_copy_area(draw->conn,
-                              buffer->pixmap,
-                              new_buffer->pixmap,
-                              dri3_drawable_gc(draw),
-                              0, 0, 0, 0,
-                              draw->width, draw->height);
-               dri3_fence_trigger(draw->conn, new_buffer);
-            } else if (draw->vtable->in_current_context(draw)) {
-               draw->ext->image->blitImage(dri_context,
-                                           new_buffer->image,
-                                           buffer->image,
-                                           0, 0, draw->width, draw->height,
-                                           0, 0, draw->width, draw->height, 0);
-            }
-            dri3_free_render_buffer(draw, buffer);
+      if ((buffer_type == loader_dri3_buffer_back ||
+           (buffer_type == loader_dri3_buffer_front && draw->have_fake_front))
+          && buffer) {
+
+         /* Fill the new buffer with data from an old buffer */
+         dri3_fence_await(draw->conn, draw, buffer);
+         if (!loader_dri3_blit_image(draw,
+                                     new_buffer->image,
+                                     buffer->image,
+                                     0, 0, draw->width, draw->height,
+                                     0, 0, 0) &&
+             !buffer->linear_buffer) {
+            dri3_fence_reset(draw->conn, new_buffer);
+            dri3_copy_area(draw->conn,
+                           buffer->pixmap,
+                           new_buffer->pixmap,
+                           dri3_drawable_gc(draw),
+                           0, 0, 0, 0,
+                           draw->width, draw->height);
+            dri3_fence_trigger(draw->conn, new_buffer);
          }
-         break;
-      case loader_dri3_buffer_front:
+         dri3_free_render_buffer(draw, buffer);
+      } else if (buffer_type == loader_dri3_buffer_front) {
+         /* Fill the new fake front with data from a real front */
          loader_dri3_swapbuffer_barrier(draw);
          dri3_fence_reset(draw->conn, new_buffer);
          dri3_copy_area(draw->conn,
@@ -1270,23 +1448,45 @@ dri3_get_buffer(__DRIdrawable *driDrawable,
                         draw->width, draw->height);
          dri3_fence_trigger(draw->conn, new_buffer);
 
-         if (new_buffer->linear_buffer &&
-             draw->vtable->in_current_context(draw)) {
-            dri3_fence_await(draw->conn, new_buffer);
-            draw->ext->image->blitImage(dri_context,
-                                        new_buffer->image,
-                                        new_buffer->linear_buffer,
-                                        0, 0, draw->width, draw->height,
-                                        0, 0, draw->width, draw->height, 0);
+         if (new_buffer->linear_buffer) {
+            dri3_fence_await(draw->conn, draw, new_buffer);
+            (void) loader_dri3_blit_image(draw,
+                                          new_buffer->image,
+                                          new_buffer->linear_buffer,
+                                          0, 0, draw->width, draw->height,
+                                          0, 0, 0);
          }
-         break;
       }
       buffer = new_buffer;
-      buffer->buffer_type = buffer_type;
       draw->buffers[buf_id] = buffer;
    }
-   dri3_fence_await(draw->conn, buffer);
+   dri3_fence_await(draw->conn, draw, buffer);
 
+   /*
+    * Do we need to preserve the content of a previous buffer?
+    *
+    * Note that this blit is needed only to avoid a wait for a buffer that
+    * is currently in the flip chain or being scanned out from. That's really
+    * a tradeoff. If we're ok with the wait we can reduce the number of back
+    * buffers to 1 for SWAP_EXCHANGE, and 1 for SWAP_COPY,
+    * but in the latter case we must disallow page-flipping.
+    */
+   if (buffer_type == loader_dri3_buffer_back &&
+       draw->cur_blit_source != -1 &&
+       draw->buffers[draw->cur_blit_source] &&
+       buffer != draw->buffers[draw->cur_blit_source]) {
+
+      struct loader_dri3_buffer *source = draw->buffers[draw->cur_blit_source];
+
+      /* Avoid flushing here. Will propably do good for tiling hardware. */
+      (void) loader_dri3_blit_image(draw,
+                                    buffer->image,
+                                    source->image,
+                                    0, 0, draw->width, draw->height,
+                                    0, 0, 0);
+      buffer->last_swap = source->last_swap;
+      draw->cur_blit_source = -1;
+   }
    /* Return the requested buffer */
    return buffer;
 }
@@ -1310,10 +1510,12 @@ dri3_free_buffers(__DRIdrawable *driDrawable,
    case loader_dri3_buffer_back:
       first_id = LOADER_DRI3_BACK_ID(0);
       n_id = LOADER_DRI3_MAX_BACK;
+      draw->cur_blit_source = -1;
       break;
    case loader_dri3_buffer_front:
       first_id = LOADER_DRI3_FRONT_ID;
-      n_id = 1;
+      /* Don't free a fake front holding new backbuffer content. */
+      n_id = (draw->cur_blit_source == LOADER_DRI3_FRONT_ID) ? 0 : 1;
    }
 
    for (buf_id = first_id; buf_id < first_id + n_id; buf_id++) {
@@ -1352,8 +1554,10 @@ loader_dri3_get_buffers(__DRIdrawable *driDrawable,
    if (!dri3_update_drawable(driDrawable, draw))
       return false;
 
-   /* pixmaps always have front buffers */
-   if (draw->is_pixmap)
+   /* pixmaps always have front buffers.
+    * Exchange swaps also mandate fake front buffers.
+    */
+   if (draw->is_pixmap || draw->swap_method == __DRI_ATTRIB_SWAP_EXCHANGE)
       buffer_mask |= __DRI_IMAGE_BUFFER_FRONT;
 
    if (buffer_mask & __DRI_IMAGE_BUFFER_FRONT) {
@@ -1430,6 +1634,7 @@ loader_dri3_update_drawable_geometry(struct loader_dri3_drawable *draw)
       draw->width = geom_reply->width;
       draw->height = geom_reply->height;
       draw->vtable->set_drawable_size(draw, draw->width, draw->height);
+      draw->ext->flush->invalidate(draw->dri_drawable);
 
       free(geom_reply);
    }
@@ -1449,3 +1654,73 @@ loader_dri3_swapbuffer_barrier(struct loader_dri3_drawable *draw)
 
    (void) loader_dri3_wait_for_sbc(draw, 0, &ust, &msc, &sbc);
 }
+
+/**
+ * Perform any cleanup associated with a close screen operation.
+ * \param dri_screen[in,out] Pointer to __DRIscreen about to be closed.
+ *
+ * This function destroys the screen's cached swap context if any.
+ */
+void
+loader_dri3_close_screen(__DRIscreen *dri_screen)
+{
+   mtx_lock(&blit_context.mtx);
+   if (blit_context.ctx && blit_context.cur_screen == dri_screen) {
+      blit_context.core->destroyContext(blit_context.ctx);
+      blit_context.ctx = NULL;
+   }
+   mtx_unlock(&blit_context.mtx);
+}
+
+/**
+ * Find a backbuffer slot - potentially allocating a back buffer
+ *
+ * \param draw[in,out]  Pointer to the drawable for which to find back.
+ * \return Pointer to a new back buffer or NULL if allocation failed or was
+ * not mandated.
+ *
+ * Find a potentially new back buffer, and if it's not been allocated yet and
+ * in addition needs initializing, then try to allocate and initialize it.
+ */
+#include <stdio.h>
+static struct loader_dri3_buffer *
+dri3_find_back_alloc(struct loader_dri3_drawable *draw)
+{
+   struct loader_dri3_buffer *back;
+   int id;
+
+   id = dri3_find_back(draw);
+   if (id < 0)
+      return NULL;
+
+   back = draw->buffers[id];
+   /* Allocate a new back if we haven't got one */
+   if (!back && draw->back_format != __DRI_IMAGE_FORMAT_NONE &&
+       dri3_update_drawable(draw->dri_drawable, draw))
+      back = dri3_alloc_render_buffer(draw, draw->back_format,
+                                      draw->width, draw->height, draw->depth);
+
+   if (!back)
+      return NULL;
+
+   draw->buffers[id] = back;
+
+   /* If necessary, prefill the back with data according to swap_method mode. */
+   if (draw->cur_blit_source != -1 &&
+       draw->buffers[draw->cur_blit_source] &&
+       back != draw->buffers[draw->cur_blit_source]) {
+      struct loader_dri3_buffer *source = draw->buffers[draw->cur_blit_source];
+
+      dri3_fence_await(draw->conn, draw, source);
+      dri3_fence_await(draw->conn, draw, back);
+      (void) loader_dri3_blit_image(draw,
+                                    back->image,
+                                    source->image,
+                                    0, 0, draw->width, draw->height,
+                                    0, 0, 0);
+      back->last_swap = source->last_swap;
+      draw->cur_blit_source = -1;
+   }
+
+   return back;
+}
